
"use strict";

var esprima = require('esprima');
var ast_visitor = require('./ast_visitor');
var ast_match = require('./ast_match');

//var FILENAME = "test.js";
//var FILENAME = "ast_match.js";
var FILENAME = "jquery-3.2.1.js";
var program = ast_visitor.readFile(FILENAME);
console.log("!!! PARSING");
var tree = esprima.parse(program);

//console.log("TOKENS: " + JSON.stringify(esprima.tokenize(top), undefined, 2));
// console.log("TREE: " + JSON.stringify(tree, undefined, 2));

// var pv = new ast_visitor.PrintVisitor();
// pv.visit(tree);

var template = ast_match.ast_to_template(tree);
//template.print();
console.log("WEIGHT: " + template.weight());

let all_subtrees = [];
template.iter_post_order((t, n) => {
    all_subtrees.push(t);
});
console.log("COUNT: " + all_subtrees.length);

/*all_subtrees.forEach((st, i) => {
    console.log("NODE " + i + ": " + st.shallow_str());
});*/

//let xt = ast_match.generate_template(all_subtrees[8], all_subtrees[9]);
//xt.print();
//console.log(xt.weight());

console.log("!!! GENERATING TEMPLATES");
let all_templates = [];
for (let i = 0; i < all_subtrees.length; i++) {
    if (i % 1000 == 0) {
        console.log("  Checked " + i + " subtrees (" + all_templates.length + " templates).");
    }
    let st1 = all_subtrees[i];
    let max_weight = -1;
    for (let j = 0; j < all_subtrees.length; j++) {
        if (i >= j) {
            continue;
        }
        let st2 = all_subtrees[j];
        let t = ast_match.generate_template(st1, st2);
        if (ast_match.isHole(t)) {
            continue;
        }
        if (t.weight() < 10) { continue; }
        all_templates.push(t);
    }
}

console.log("Generated " + all_templates.length + " templates.");

// Go through all possible templates, and hang a list of all of the subtrees that match it, and their
// weights, off of it.
console.log("!!! DEDUP TEMPLATES");
let dedup_templates = [];
for (let i = 0; i < all_templates.length; i++) {
    if (i % 10000 == 0) {
        console.log("Finished " + i + "/" + all_templates.length + " - got " + dedup_templates.length);
    }
    let t = all_templates[i];
    let exists = false;
    for (let j = 0; j < dedup_templates.length; j++) {
        if (t.equals(dedup_templates[j])) {
            exists = true;
            break;
        }
    }
    if (exists)
        continue;
    dedup_templates.push(t);
}

// Go through the deduped templates, and associate them with all the subtrees they match.
// weights, off of it.
console.log("!!! MATCH TEMPLATES");
for (let i = 0; i < dedup_templates.length; i++) {
    if (i % 10 == 0) {
        console.log("Done " + i + "/" + dedup_templates.length);
    }
    let t = dedup_templates[i];
    t.match_list = [];
    for (let j = 0; j < all_subtrees.length; j++) {
        let st = all_subtrees[j];
        if (t.matches(st)) {
            t.match_list.push(st);
        }
    }
}

// Sort the deduped templates by weight*match_list
console.log("!!! SORT TEMPLATES");
dedup_templates.sort((a, b) => {
    return -((b.weight() * b.match_list.length) - (a.weight() * a.match_list.length));
});

// Print it all out.
dedup_templates.forEach((t, i) => {
    console.log("Template " + t.shallow_str());
    let value = t.weight() * t.match_list.length;
    console.log("   " + value + " w=" + t.weight() + " n=" + t.match_list.length);
    t.print(2);
});

console.log("Deduped-templates: " + dedup_templates.length);

console.log("!!! DONE");
